<?php

declare(strict_types=1);

namespace Drupal\entity_usage_updater;

use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityPublishedInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Entity\RevisionableInterface;
use Drupal\Core\Entity\RevisionableStorageInterface;
use Drupal\Core\Entity\RevisionLogInterface;
use Drupal\Core\Entity\TranslatableInterface;
use Drupal\Core\Entity\TranslatableRevisionableStorageInterface;
use Drupal\Core\Field\FieldItemInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\entity_reference_revisions\EntityNeedsSaveInterface;
use Drupal\entity_usage\EntityUsageInterface;
use Drupal\entity_usage\EntityUsageTrackInterface;
use Drupal\entity_usage_updater\Utility\ViolationsHelper;
use Drupal\paragraphs\ParagraphInterface;
use Drupal\redirect\Entity\Redirect;

/**
 * Updates references to entities tracked with Entity Usage.
 *
 * @todo Add lock.
 * @todo Support optional entity validation prior to saving.
 * @todo Check that ERR fields aren't translatable (which we assume).
 * @todo check that content moderation isn't enabled for entities being updated?
 * @todo Sanitize exceptions.
 * @todo Improve details in error logging.
 */
class EntityUsageUpdater implements EntityUsageUpdaterInterface {

  /**
   * {@inheritdoc}
   */
  public function update(array $mapping): void {
    $batch_builder = new BatchBuilder();
    $batch_builder
      ->setFinishCallback([static::class, 'finish'])
      ->setTitle(t('Entity reference update'))
      ->setInitMessage(t('Updating entity references.'))
      ->setErrorMessage(t('There was an error updating entity references.'));

    foreach ($mapping as $old_entity_type_id => $entity_updates) {
      foreach ($entity_updates as $old_id => [$new_entity_type_id, $new_id]) {
        $arguments = [
          $old_entity_type_id,
          $old_id,
          $new_entity_type_id,
          $new_id,
        ];
        $batch_builder->addOperation([static::class, 'createBatchOperationsFromUsages'], $arguments);
      }
    }
    $batch_builder->addOperation([static::class, 'triggerBatch'], [[]]);

    batch_set($batch_builder->toArray());
  }

  /**
   * {@inheritdoc}
   */
  public function remove(array $mapping, array $moderation = []): void {
    $batch_builder = new BatchBuilder();
    $batch_builder
      ->setFinishCallback([static::class, 'finish'])
      ->setTitle(t('Entity reference remove'))
      ->setInitMessage(t('Creating batch operations to remove entity references.'))
      ->setErrorMessage(t('There was an error removing entity references.'));

    foreach ($mapping as $old_entity_type_id => $entity_updates) {
      foreach ($entity_updates as $old_id) {
        $arguments = [
          $old_entity_type_id,
          $old_id,
          $moderation[$old_entity_type_id][$old_id] ?? NULL,
        ];
        $batch_builder->addOperation([static::class, 'createRemoveBatchOperationsFromUsages'], $arguments);
      }
    }

    $batch_builder->addOperation([static::class, 'addRemoveOperationsForRedirects']);
    $batch_builder->addOperation([static::class, 'triggerBatch'], [$mapping]);

    batch_set($batch_builder->toArray());
  }

  /**
   * Finds usages of a particular entity and adds batch update operations.
   *
   * It exists as a standalone batch operation in case the usages of a single
   * entity are particularly large.
   *
   * @param string $old_entity_type_id
   *   The entity type of the entity we're searching for.
   * @param string|int $old_id
   *   The entity ID of the entity we're searching for.
   * @param string $new_entity_type_id
   *   The entity type of the entity to replace it with.
   * @param string|int $new_id
   *   The entity ID of the entity to replace it with.
   * @param array &$context
   *   The batch context.
   */
  public static function createBatchOperationsFromUsages(string $old_entity_type_id, $old_id, string $new_entity_type_id, $new_id, array &$context): void {
    $entity = \Drupal::entityTypeManager()->getStorage($old_entity_type_id)->load($old_id);
    if (!$entity) {
      $context['results']['errors'][] = t("Couldn't load @entity_type @id", [
        '@entity_type_id' => $old_entity_type_id,
        '@id' => $old_id,
      ]);
      return;
    }

    $usages = self::getUsages($entity);
    if (empty($usages)) {
      return;
    }

    if (!isset($context['results']['batch'])) {
      $batch_builder = new BatchBuilder();
      $batch_builder
        ->setFinishCallback([static::class, 'finish'])
        ->setTitle(t('Entity reference update'))
        ->setInitMessage(t('Updating entity references.'))
        ->setErrorMessage(t('There was an error updating entity references.'));
    }
    else {
      $batch_builder = $context['results']['batch'];
    }

    // Paragraph integration.
    if (isset($usages['paragraph'])) {
      foreach ($usages['paragraph'] as $entity_id => $entity_usages) {
        $arguments = [
          $entity_id,
          $entity_usages,
        ];
        $batch_builder->addOperation([static::class, 'addHostEntityToBatch'], $arguments);
      }
    }

    foreach ($usages as $entity_type_id => $entity_type_usages) {
      // Paragraphs are handled as part of the host entity.
      if ($entity_type_id === 'paragraph') {
        continue;
      }
      foreach ($entity_type_usages as $entity_id => $entity_usages) {
        $arguments = [
          $entity,
          $new_entity_type_id,
          $new_id,
          $entity_type_id,
          $entity_id,
          $entity_usages,
          'processReferences',
        ];
        $batch_builder->addOperation([static::class, 'updateReferences'], $arguments);
      }
    }

    $arguments = [
      $entity,
      $new_entity_type_id,
      $new_id,
      '',
      '',
      [],
      'processHostEntities',
    ];
    $batch_builder->addOperation([static::class, 'updateReferences'], $arguments);
    $context['results']['batch'] = $batch_builder;
    $context['message'] = t('Added batch operations for %label.', ['%label' => $entity->label() ?? $entity->getEntityTypeId() . ':' . $entity->id()]);
  }

  /**
   * Finds usages of a particular entity and adds batch remove operations.
   *
   * It exists as a standalone batch operation in case the usages of a single
   * entity are particularly large.
   *
   * @param string $old_entity_type_id
   *   The entity type of the entity we're searching for.
   * @param string|int $old_id
   *   The entity ID of the entity we're searching for.
   * @param string|null $moderation_state
   *   The publishing state to change the entity to after removing usages.
   *   Supports content moderation states.
   * @param array &$context
   *   The batch context.
   */
  public static function createRemoveBatchOperationsFromUsages(string $old_entity_type_id, $old_id, string|null $moderation_state, array &$context): void {
    $entity = \Drupal::entityTypeManager()->getStorage($old_entity_type_id)->load($old_id);
    if (!$entity) {
      $context['results']['errors'][] = t("Couldn't load @entity_type @id", [
        '@entity_type_id' => $old_entity_type_id,
        '@id' => $old_id,
      ]);
      return;
    }

    $usages = self::getUsages($entity);
    if (empty($usages) && $moderation_state === NULL) {
      // There is nothing to do.
      return;
    }

    if (!isset($context['results']['batch'])) {
      $batch_builder = new BatchBuilder();
      $batch_builder
        ->setFinishCallback([static::class, 'finish'])
        ->setTitle(t('Entity reference update'))
        ->setInitMessage(t('Removing entity references.'))
        ->setErrorMessage(t('There was an error removing entity references.'));
    }
    else {
      $batch_builder = $context['results']['batch'];
    }

    // Redirect integration.
    if (isset($usages['redirect'])) {
      $context['results']['redirects'] = array_unique(array_merge($context['results']['redirects'] ?? [], array_keys($usages['redirect'])));
    }

    // Paragraph integration.
    if (isset($usages['paragraph'])) {
      foreach ($usages['paragraph'] as $entity_id => $entity_usages) {
        $arguments = [
          $entity_id,
          $entity_usages,
        ];
        $batch_builder->addOperation([static::class, 'addHostEntityToBatch'], $arguments);
      }
    }

    foreach ($usages as $entity_type_id => $entity_type_usages) {
      // Redirects do not have a field UI and the only trackable field on it is
      // a link base field that is required so we cannot remove the value
      // without creating an validation error. If the redirect points to an
      // entity you are removing links to and unpublishing then the resulting
      // redirect will still work and redirect people to the now unpublished
      // entity in the same way if they had the direct link to content.
      if ($entity_type_id === 'redirect') {
        continue;
      }
      // Paragraphs are handled as part of the host entity.
      if ($entity_type_id === 'paragraph') {
        continue;
      }
      foreach ($entity_type_usages as $entity_id => $entity_usages) {
        $arguments = [
          $entity,
          $entity_type_id,
          $entity_id,
          $entity_usages,
          'processReferences',
        ];
        $batch_builder->addOperation([
          static::class,
          'removeReferences',
        ], $arguments);
      }
    }

    if (!empty($usages)) {
      $arguments = [
        $entity,
        '',
        '',
        [],
        'processHostEntities',
      ];
      $batch_builder->addOperation([
        static::class,
        'removeReferences',
      ], $arguments);
    }

    if ($moderation_state !== NULL) {
      $batch_builder->addOperation([static::class, 'moderateEntity'], [$old_entity_type_id, $old_id, $moderation_state]);
    }

    $context['results']['batch'] = $batch_builder;
    $context['message'] = t('Added batch operations for %label.', ['%label' => $entity->label() ?? $entity->getEntityTypeId() . ':' . $entity->id()]);
  }

  /**
   * Adds operations to the remove batch to process redirects.
   *
   * @param array $context
   *   The batch context.
   *
   * @return void
   */
  public static function addRemoveOperationsForRedirects(array &$context): void {
    if (empty($context['sandbox'])) {
      if (empty($context['results']['redirects'])) {
        return;
      }
      $context['sandbox']['redirects'] = $context['results']['redirects'];
      $context['sandbox']['position'] = 0;
      unset($context['results']['redirects']);
    }

    $entity_id = $context['sandbox']['redirects'][$context['sandbox']['position']];
    static::createRemoveBatchOperationsFromUsages('redirect', $entity_id, NULL, $context);

    if (isset($context['results']['redirects'])) {
      // Add an new redirects to the list. It is possible to configure a
      // redirect loop.
      $new_redirects = array_diff($context['results']['redirects'], $context['sandbox']['redirects']);
      $context['sandbox']['redirects'] = array_merge($context['sandbox']['redirects'], $new_redirects);
    }
    $context['sandbox']['position']++;
    $context['message'] = t('Adding operations for redirects');
    $context['finished'] = $context['sandbox']['position'] / count($context['sandbox']['redirects']);
  }

  /**
   * Updates references within a single entity.
   *
   * It handles all translations for the given entity in one go.
   *
   * @param \Drupal\Core\Entity\EntityInterface $old_target
   *   The entity that should no longer be referenced.
   * @param string $new_target_entity_type
   *   The entity type of the entity that should be referenced instead of
   *   $old_target.
   * @param int|string $new_target_id
   *   The entity ID of the entity that should be referenced instead of
   *   $old_target.
   * @param string $entity_type_id
   *   The entity type of the entity being updated.
   * @param int|string $entity_id
   *   The ID of the entity being updated.
   * @param string[][] $usages
   *   An array of associative arrays of usage information, keyed by:
   *   - source_langcode
   *   - method
   *   - field_name
   *   - count.
   * @param string $method
   *   The method to call to process the updates.
   * @param array &$context
   *   The batch API context array.
   */
  public static function updateReferences(EntityInterface $old_target, string $new_target_entity_type, $new_target_id, string $entity_type_id, $entity_id, array $usages, string $method, array &$context) {
    $log_message = (string) t("Automatically updated by the entity usage updater module to replace @old_entity_type @old_entity_id with @new_entity_type @new_entity_id", [
      '@old_entity_type' => $old_target->getEntityTypeId(),
      '@old_entity_id' => $old_target->id(),
      '@new_entity_type' => $new_target_entity_type,
      '@new_entity_id' => $new_target_id,
    ]);
    $func = function ($updater, $item) use ($old_target, $new_target_entity_type, $new_target_id) {
      $updater->update($old_target, $new_target_entity_type, $new_target_id, $item);
    };
    // @phpstan-ignore match.unhandled
    match ($method) {
      'processReferences' => self::processReferences($old_target, $entity_type_id, $entity_id, $func, $log_message, $usages, $context),
      'processHostEntities' => self::processHostEntities($old_target, $func, $log_message, $context),
    };
  }

  /**
   * Removes references within a single entity.
   *
   * It handles all translations for the given entity in one go.
   *
   * @param \Drupal\Core\Entity\EntityInterface $old_target
   *   The entity that should no longer be referenced.
   * @param string $entity_type_id
   *   The entity type of the entity being updated.
   * @param int|string $entity_id
   *   The ID of the entity being updated.
   * @param string[][] $usages
   *   An array of associative arrays of usage information, keyed by:
   *   - source_langcode
   *   - method
   *   - field_name
   *   - count.
   * @param string $method
   *   The method to call to process the removals.
   * @param array &$context
   *   The batch API context array.
   */
  public static function removeReferences(EntityInterface $old_target, string $entity_type_id, $entity_id, array $usages, string $method, array &$context): void {
    $func = function (EntityUsageUpdaterPluginInterface $updater, FieldItemInterface $item) use ($old_target) {
      $updater->remove($old_target, $item);
    };
    $log_message = (string) t("Automatically updated by the entity usage updater module to remove @old_entity_type @old_entity_id", [
      '@old_entity_type' => $old_target->getEntityTypeId(),
      '@old_entity_id' => $old_target->id(),
    ]);
    $context['sandbox']['filter_empty_items'] = TRUE;
    // @phpstan-ignore match.unhandled
    match ($method) {
      'processReferences' => self::processReferences($old_target, $entity_type_id, $entity_id, $func, $log_message, $usages, $context),
      'processHostEntities' => self::processHostEntities($old_target, $func, $log_message, $context),
    };
  }

  /**
   * Processes references within a single entity.
   *
   * It handles all translations for the given entity in one go.
   *
   * @param \Drupal\Core\Entity\EntityInterface $old_target
   *   The entity that should no longer be referenced.
   * @param string $entity_type_id
   *   The entity type of the entity being updated.
   * @param int|string $entity_id
   *   The ID of the entity being updated.
   * @param callable $func
   *   A function that accepts an EntityUsageUpdaterPluginInterface and a field
   *   item to update.
   * @param string $log_message
   *   The log message to use when the entity is saved.
   * @param string[][] $usages
   *   An array of associative arrays of usage information, keyed by:
   *   - source_langcode
   *   - method
   *   - field_name
   *   - count.
   * @param array &$context
   *   The batch API context array.
   */
  private static function processReferences(EntityInterface $old_target, string $entity_type_id, $entity_id, callable $func, string $log_message, array $usages, array &$context): void {
    $storage = \Drupal::entityTypeManager()->getStorage($entity_type_id);
    $entity_type = \Drupal::entityTypeManager()->getDefinition($entity_type_id);
    $updater_manager = \Drupal::service('plugin.manager.entity_usage_updater');
    assert($updater_manager instanceof PluginManagerInterface);
    $tracker_manager = \Drupal::service('plugin.manager.entity_usage.track');
    assert($tracker_manager instanceof PluginManagerInterface);

    $current_revision = $storage->load($entity_id);
    // We should only change the latest revision. I.e if there are drafts we can
    // only update the latest draft. Changing both the published revision and
    // future draft is impossible and would need an editor to decide how to
    // manage this.
    if ($storage->getEntityType()->isRevisionable()) {
      assert($storage instanceof RevisionableStorageInterface);
      assert($current_revision instanceof RevisionableInterface);
      $vid = $storage->getLatestRevisionId($entity_id);
      if ($current_revision->getRevisionId() != $vid) {
        $current_revision = $storage->loadRevision($vid);
        $context['results']['draft_revisions'][$current_revision->getEntityTypeId()][] = $current_revision->id();
      }
    }

    // Determine if the entity is currently valid.
    assert($current_revision instanceof FieldableEntityInterface);
    $violations = $current_revision->validate();
    $initial_entity_violations = $violations->count();

    if ($entity_type->isRevisionable()) {
      $new_revision = static::createRevision($current_revision);
    }
    else {
      $new_revision = $current_revision;
    }

    foreach ($usages as $reference) {
      $langcode = $reference['source_langcode'];

      if ($current_revision instanceof TranslatableInterface) {
        assert($new_revision instanceof TranslatableInterface);
        $new_translation = $new_revision->getTranslation($langcode);
      }
      else {
        $new_translation = $new_revision;
      }
      assert($new_translation instanceof FieldableEntityInterface);
      assert(!$new_translation instanceof RevisionableInterface || $new_translation->isLatestRevision());
      assert($new_translation->language()->getId() === $reference['source_langcode']);

      $method = $reference['method'];
      try {
        // Use the Entity Usage tracker to help us identify which field deltas
        // contain references to the target entity. (That level of detail isn't
        // provided in the usage information.)
        $tracker = $tracker_manager->createInstance($method);
        assert($tracker instanceof EntityUsageTrackInterface);
        $updater = $updater_manager->createInstance($method);
        assert($updater instanceof EntityUsageUpdaterPluginInterface);

        $field = $new_translation->get($reference['field_name']);
        assert($field instanceof FieldItemListInterface);
        foreach ($field as $item) {
          // Use entity_usage to verify if the target entity is in the field.
          $old_target_string = $old_target->getEntityTypeId() . '|' . $old_target->id();
          if (in_array($old_target_string, $tracker->getTargetEntities($item), TRUE)) {
            $func($updater, $item);
          }
        }
        if ($context['sandbox']['filter_empty_items'] ?? FALSE) {
          $field->filterEmptyItems();
        }
      }
      catch (PluginException $e) {
        $context['results']['errors'][] = $e->getMessage();
      }
    }

    $paragraph_key = $current_revision->getEntityTypeId() . ':' . $current_revision->id();
    if (!empty($context['results']['entities_with_paragraphs_to_update'][$paragraph_key])) {
      self::processParagraphs($current_revision, $context['results']['entities_with_paragraphs_to_update'][$paragraph_key], $tracker_manager, $updater_manager, $old_target, $func, $context);
      unset($context['results']['entities_with_paragraphs_to_update'][$paragraph_key]);
    }

    if ($new_revision instanceof RevisionLogInterface) {
      $new_revision->setRevisionLogMessage($log_message);
    }

    // Ensure we are not making increasing the number of violations.
    assert($new_revision instanceof FieldableEntityInterface);
    $violations = $new_revision->validate();
    if ($violations->count() > $initial_entity_violations) {
      $context['results']['errors'][] = ViolationsHelper::toTranslatableMarkup($violations);
      \Drupal::logger('entity_usage_updater')->error(...ViolationsHelper::toLogMessage($violations));
      return;
    }

    $new_revision->save();

    $key = $old_target->getEntityTypeId() . ':' . $old_target->id();
    if (!isset($context['results']['update_count'][$key])) {
      $context['results']['update_count'][$key] = 0;
    }
    $context['results']['update_count'][$key]++;
    $context['message'] = t('Processing references for %label.', ['%label' => $new_revision->label() ?? $new_revision->getEntityTypeId() . ':' . $new_revision->id()]);
  }

  /**
   * Processes paragraphs within a single entity.
   *
   * It handles all translations and paragraphs for the given entity in one go.
   *
   * @param \Drupal\Core\Entity\EntityInterface $old_target
   *   The entity that should no longer be referenced.
   * @param callable $func
   *   A function that accepts an EntityUsageUpdaterPluginInterface and a field
   *   item to update.
   * @param string $log_message
   *   The log message to use when the entity is saved.
   * @param array &$context
   *   The batch API context array.
   */
  protected static function processHostEntities(EntityInterface $old_target, callable $func, string $log_message, array &$context): void {
    // Use sandbox for communication between iterations of the same operation.
    if (!empty($context['results']['entities_with_paragraphs_to_update'])) {
      $context['sandbox']['entities_with_paragraphs_to_update'] = $context['results']['entities_with_paragraphs_to_update'];
      $context['results']['entities_with_paragraphs_to_update'] = [];
      $context['sandbox']['total'] = count($context['sandbox']['entities_with_paragraphs_to_update']);
      $context['sandbox']['count'] = 0;
    }

    // Return if we've already handled all entities with updated ERR field
    // targets.
    if (empty($context['sandbox']['entities_with_paragraphs_to_update'])) {
      $context['finished'] = 1;
      return;
    }

    $updater_manager = \Drupal::service('plugin.manager.entity_usage_updater');
    assert($updater_manager instanceof PluginManagerInterface);
    $tracker_manager = \Drupal::service('plugin.manager.entity_usage.track');
    assert($tracker_manager instanceof PluginManagerInterface);

    $data = array_splice($context['sandbox']['entities_with_paragraphs_to_update'], 0, 1);
    [$entity_type_id, $id] = explode(':', key($data), 2);

    $storage = \Drupal::entityTypeManager()->getStorage($entity_type_id);
    $host_entity = $storage->load($id);
    // We should only change the latest revision. I.e if there are drafts we can
    // only update the latest draft. Changing both the published revision and
    // future draft is impossible and would need an editor to decide how to
    // manage this.
    if ($storage->getEntityType()->isRevisionable()) {
      assert($storage instanceof RevisionableStorageInterface);
      assert($host_entity instanceof RevisionableInterface);
      $vid = $storage->getLatestRevisionId($id);
      if ($host_entity->getRevisionId() != $vid) {
        $host_entity = $storage->loadRevision($vid);
        $context['results']['draft_revisions'][$host_entity->getEntityTypeId()][] = $host_entity->id();
      }
    }

    // Determine if the entity is currently valid.
    assert($host_entity instanceof FieldableEntityInterface);
    $violations = $host_entity->validate();
    $initial_entity_violations = $violations->count();

    if ($host_entity->getEntityType()->isRevisionable()) {
      /** @var \Drupal\Core\Entity\FieldableEntityInterface $host_entity */
      $host_entity = static::createRevision($host_entity);
    }
    self::processParagraphs($host_entity, reset($data), $tracker_manager, $updater_manager, $old_target, $func, $context);

    if ($host_entity instanceof RevisionLogInterface) {
      $host_entity->setRevisionLogMessage($log_message);
    }

    // Ensure we are not making increasing the number of violations.
    $violations = $host_entity->validate();
    if ($violations->count() > $initial_entity_violations) {
      $context['results']['errors'][] = ViolationsHelper::toTranslatableMarkup($violations);
      \Drupal::logger('entity_usage_updater')->error(...ViolationsHelper::toLogMessage($violations));
      return;
    }

    $host_entity->save();

    $key = $old_target->getEntityTypeId() . ':' . $old_target->id();
    if (!isset($context['results']['update_count'][$key])) {
      $context['results']['update_count'][$key] = 0;
    }
    $context['results']['update_count'][$key]++;
    $context['sandbox']['count']++;
    $context['finished'] = $context['sandbox']['count'] / $context['sandbox']['total'];
    $context['message'] = t('Processing paragraphs for %label.', ['%label' => $host_entity->label() ?? $host_entity->getEntityTypeId() . ':' . $host_entity->id()]);
  }

  /**
   * Updated references in paragraphs on a host entity.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $host_entity
   *   The host entity to update.
   * @param array $paragraph_data
   *   The paragraph data created by ::addHostEntityToBatch()
   * @param \Drupal\Component\Plugin\PluginManagerInterface $tracker_manager
   *   The entity usage tracker plugin manager.
   * @param \Drupal\Component\Plugin\PluginManagerInterface $updater_manager
   *   The entity usage updater plugin manager.
   * @param \Drupal\Core\Entity\EntityInterface $old_target
   *   The entity that should no longer be referenced.
   * @param callable $func
   *   A function that accepts an EntityUsageUpdaterPluginInterface and a field
   *   item to update.
   * @param array $context
   *   The batch API context array.
   */
  private static function processParagraphs(FieldableEntityInterface $host_entity, array $paragraph_data, PluginManagerInterface $tracker_manager, PluginManagerInterface $updater_manager, EntityInterface $old_target, callable $func, array &$context): void {
    foreach ($paragraph_data as $langcode => $references) {
      if ($host_entity instanceof TranslatableInterface) {
        $host_entity = $host_entity->getTranslation($langcode);
      }
      foreach ($references as $reference) {
        $entity_to_update = $host_entity;
        $found = FALSE;
        foreach ($reference as $paragraph) {
          $field_item_list = $entity_to_update->get($paragraph['field_name']);
          $found = FALSE;
          foreach ($field_item_list as $field_item) {
            if ($field_item->entity->id() === $paragraph['id']) {
              $entity_to_update = $field_item->entity;
              // We need to mark the entity as needing a save so that changes
              // are saved to nested paragraphs.
              if ($entity_to_update instanceof EntityNeedsSaveInterface) {
                $entity_to_update->setNeedsSave(TRUE);
              }
              $found = TRUE;
              continue 2;
            }
          }
          // @phpstan-ignore booleanNot.alwaysTrue
          if (!$found) {
            // This should not happen. Orphaned paragraphs are removed by
            // ::getUsages().
            continue 2;
          }
        }
        if ($found && isset($paragraph['usages'])) {
          foreach ($paragraph['usages'] as $usage) {
            $method = $usage['method'];
            try {
              // Use the Entity Usage tracker to help us identify which field deltas
              // contain references to the target entity. (That level of detail isn't
              // provided in the usage information.)
              $tracker = $tracker_manager->createInstance($method);
              assert($tracker instanceof EntityUsageTrackInterface);
              $updater = $updater_manager->createInstance($method);
              assert($updater instanceof EntityUsageUpdaterPluginInterface);

              $field = $entity_to_update->get($usage['field_name']);
              assert($field instanceof FieldItemListInterface);
              foreach ($field as $item) {
                // Use entity_usage to verify if the target entity is in the field.
                $old_target_string = $old_target->getEntityTypeId() . '|' . $old_target->id();
                if (in_array($old_target_string, $tracker->getTargetEntities($item), TRUE)) {
                  $func($updater, $item);
                }
              }
              if ($context['sandbox']['filter_empty_items'] ?? FALSE) {
                $field->filterEmptyItems();
              }
            }
            catch (PluginException $e) {
              $context['results']['errors'][] = $e->getMessage();
            }
          }
        }
      }
    }
  }

  /**
   * Adds a host entity to the batch if necessary.
   *
   * @param string $entity_id
   *   The entity ID being processed.
   * @param array $usages
   *   The usages being processed.
   * @param array $context
   *   The batch context.
   */
  public static function addHostEntityToBatch(string $entity_id, array $usages, array &$context): void {
    $storage = \Drupal::entityTypeManager()->getStorage('paragraph');
    assert($storage instanceof RevisionableStorageInterface);
    $vid = $storage->getLatestRevisionId($entity_id);
    $entity = $storage->loadRevision($vid);
    // The paragraph must have been deleted.
    if (!$entity) {
      return;
    }

    $paragraphs = [];
    $langcode = reset($usages)['source_langcode'];
    // @todo More generally we want to consider any entity that's referenced by
    //   an ERR field. However I don't know if I can make the same assumptions
    //   about i18n with ERR fields in general that I can with paragraphs
    //   (because the paragraphs instructions are very clear).
    while ($entity->getEntityTypeId() === 'paragraph') {
      assert($entity instanceof ParagraphInterface);
      $paragraphs[] = [
        'id' => $entity->id(),
        'field_name' => $entity->get('parent_field_name')->value,
        'usages' => $usages,
      ];
      $usages = NULL;
      $entity = $entity->getParentEntity();
    }

    if (!empty($paragraphs)) {
      $key = $entity->getEntityTypeId() . ':' . $entity->id();
      // If we assume everything works like paragraphs wrt i18n then the
      // language is irrelevant. The ERR field is not translatable, and its
      // one value should be updated to the latest paragraph vid. That in turn
      // should have all the correct values for the various languages already.
      $context['results']['entities_with_paragraphs_to_update'][$key][$langcode][] = array_reverse($paragraphs);
      $context['message'] = t('%label has paragraphs to process', ['%label' => $entity->label() ?? $entity->getEntityTypeId() . ':' . $entity->id()]);
    }
  }

  /**
   * Returns the active revision ID for a given entity.
   *
   * The active revision is appropriate for editing, though it might not be the
   * currently visible one. I'm not sure this plays nicely with all possible
   * forms of multilingual pending revisions.
   *
   * @param string $entity_type_id
   *   The entity type.
   * @param int|string $entity_id
   *   The entity ID.
   * @param string $langcode
   *   The language code.
   *
   * @return int
   *   The revision ID.
   *
   * @throws \RuntimeException
   *   If no revisions are found.
   */
  protected static function getActiveRevisionId(string $entity_type_id, $entity_id, string $langcode): int {
    $storage = \Drupal::entityTypeManager()->getStorage($entity_type_id);
    // Latest revision IDs are cached by ContentEntityStorageBase.
    assert($storage instanceof RevisionableStorageInterface);
    if ($storage instanceof TranslatableRevisionableStorageInterface) {
      $active_vid = $storage->getLatestTranslationAffectedRevisionId($entity_id, $langcode);
    }
    else {
      $active_vid = $storage->getLatestRevisionId($entity_id);
    }
    if ($active_vid === NULL) {
      throw new \RuntimeException("No revision found for $entity_type_id $entity_id $langcode");
    }
    return intval($active_vid);
  }

  /**
   * Creates a new revision from an entity/revision if possible.
   *
   * If a new revision is created it will be default if $entity was default.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to create a new revision of.
   *
   * @return \Drupal\Core\Entity\EntityInterface
   *   For revisionable entities a new revision; otherwise $entity.
   */
  protected static function createRevision(EntityInterface $entity): EntityInterface {
    $storage = \Drupal::entityTypeManager()->getStorage($entity->getEntityTypeId());
    if ($storage->getEntityType()->isRevisionable()) {
      assert($entity instanceof RevisionableInterface);
      assert($storage instanceof RevisionableStorageInterface);
      return $storage->createRevision($entity, $entity->isDefaultRevision());
    }
    return $entity;
  }

  /**
   * Triggers the batch created to remove or update entity usages.
   *
   * @param array $check_usages
   *   The list of entities that are supposed to be removed. This value will be
   *   empty for the update use-case.
   * @param array $context
   *   The batch context.
   */
  public static function triggerBatch(array $check_usages, array &$context): void {
    if (isset($context['results']['batch'])) {
      $batch = $context['results']['batch'];
      assert($batch instanceof BatchBuilder);
      if (!empty($check_usages)) {
        $batch->addOperation([static::class, 'checkUsageAfterRemove'], [$check_usages]);
      }
      batch_set($batch->toArray());
      unset($context['results']['batch']);
    }
  }

  /**
   * Checks that all usages have actually been removed.
   *
   * @param array $check_usages
   *   The list of entities that are supposed to be removed.
   * @param array $context
   *   The batch context.
   *
   * @return void
   */
  public static function checkUsageAfterRemove(array $check_usages, array &$context): void {
    foreach ($check_usages as $old_entity_type_id => $entity_updates) {
      foreach (\Drupal::entityTypeManager()->getStorage($old_entity_type_id)->loadMultiple($entity_updates) as $entity) {
        if (static::countUsages($entity) > 0) {
          $context['results']['errors'][] = t(
            'There are still usages for %label (@url).',
            [
              '%label' => $entity->label() ?? $entity->getEntityTypeId() . ':' . $entity->id(),
              '@url' => $entity->toUrl()->toString(),
            ]
          );
        }
      }
    }
  }

  /**
   * Counts the number of usages of an entity's links.
   *
   * This follows redirects and counts them too as redirects are not changed
   * when you remove links.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to count usages for.
   * @param array $seen_redirects
   *   (internal) The redirects accounted for so far. Used to prevent loops.
   *
   * @return int
   *   The number of active usages.
   */
  public static function countUsages(EntityInterface $entity, array $seen_redirects = []): int {
    $usage_count = 0;
    foreach (EntityUsageUpdater::getUsages($entity) as $entity_type => $usages) {
      if ($entity_type === 'redirect') {
        foreach (Redirect::loadMultiple(array_keys($usages)) as $redirect) {
          if (!in_array($redirect->id(), $seen_redirects, TRUE)) {
            $seen_redirects[] = $redirect->id();
            $usage_count += static::countUsages($redirect, $seen_redirects);
          }
        }
      }
      else {
        $usage_count += count($usages);
      }
    }
    return $usage_count;
  }

  /**
   * Displays messages after finishing with a batch.
   *
   * It's fairly generic and used for different batches.
   *
   * @param bool $success
   *   Indicate that the batch API tasks were all completed successfully.
   * @param array $results
   *   An array of all the results that were updated in update_do_one().
   * @param array $operations
   *   A list of the operations that had not been completed by the batch API.
   */
  public static function finish($success, $results, $operations) {
    if (!$success) {
      \Drupal::messenger()->addError("The update didn't complete successfully.");
    }

    $map = [
      'errors' => MessengerInterface::TYPE_ERROR,
      'warnings' => MessengerInterface::TYPE_WARNING,
      'messages' => MessengerInterface::TYPE_STATUS,
    ];
    foreach ($map as $key => $type) {
      if (!empty($results[$key])) {
        foreach ($results[$key] as $message) {
          \Drupal::messenger()->addMessage($message, $type);
        }
      }
    }

    if ($success && !empty($results['update_count'])) {
      foreach ($results['update_count'] as $key => $count) {
        [$entity_type_id, $entity_id] = explode(':', $key, 2);
        $for_entity = \Drupal::entityTypeManager()->getStorage($entity_type_id)->load($entity_id);
        $message = \Drupal::translation()->formatPlural(
          $count,
          "Updated one entity for references to %label.",
          "Updated @count entities for references to %label.",
          ['%label' => $for_entity?->label() ?? $key]
        );
        \Drupal::messenger()->addStatus($message);
      }
    }
    if ($success && !empty($results['draft_revisions'])) {
      foreach ($results['draft_revisions'] as $entity_type => $entities) {
        foreach (\Drupal::entityTypeManager()->getStorage($entity_type)->loadMultiple($entities) as $entity) {
          \Drupal::messenger()->addWarning(
            \Drupal::translation()->translate(
              'The latest revision of <a title="Edit @title" href="@link">@title</a> has been updated. The published version is unchanged.',
              [
                '@type' => $entity->getEntityType()->getLabel(),
                '@title' => $entity->label(),
                '@link' => $entity->toUrl('edit-form')->toString(),
              ]
            )
          );
        }
      }
    }
    if ($success && isset($results['moderation_message'])) {
      \Drupal::messenger()->addStatus($results['moderation_message']);
    }
  }

  /**
   * Gets the list of usages for an entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to get usages of.
   *
   * @return array<string, array<string, array<array{source_langcode: string, method: string, field_name: string, count: string}>>>
   *   See \Drupal\entity_usage\EntityUsageInterface::listSources().
   *
   * @todo Move this to a service.
   */
  public static function getUsages(EntityInterface $entity): array {
    $entity_usage = \Drupal::service('entity_usage.usage');
    assert($entity_usage instanceof EntityUsageInterface);
    $usages = $entity_usage->listSources($entity);
    foreach ($usages as $entity_type_id => $entity_type_usages) {
      // For revisionable entity types, we're only interested in references that
      // are in the active revision.
      if (\Drupal::entityTypeManager()->getDefinition($entity_type_id)->isRevisionable()) {
        foreach ($entity_type_usages as $entity_id => $entity_usages) {
          $is_active = fn(array $usage): bool => self::isActive($entity_type_id, $entity_id, $usage['source_vid'], $usage['source_langcode']);
          $usages[$entity_type_id][$entity_id] = array_values(array_filter($entity_usages, $is_active));
          // Unset the source VID - we've already verified it's the latest
          // revision. It's possible the recorded source VID might no longer be
          // the latest one when we come to update the usage, so keeping it can
          // be actively misleading - just always update the latest revision in
          // the appropriate language.
          foreach ($usages[$entity_type_id][$entity_id] as $key => $batch_item) {
            unset($usages[$entity_type_id][$entity_id][$key]['source_vid']);
          }
        }
      }
      // Remove any empty arrays (ie entities that only had references in old
      // revisions).
      $usages[$entity_type_id] = array_filter($usages[$entity_type_id]);
    }
    // Remove any empty array (ie. entity types that are now empty).
    return array_filter($usages);
  }

  /**
   * Determines if the usage is active.
   *
   * Active is this sense means as the result of the latest revision or in the
   * case of paragraphs attached to non-revisionable entities that the paragraph
   * is the latest revision and not orphaned.
   *
   * @param string $entity_type_id
   *   The entity type ID.
   * @param string|int $entity_id
   *   The entity ID.
   * @param string|int $entity_vid
   *   The entity revision ID.
   * @param string $langcode
   *   The langcode.
   *
   * @return bool
   *   TRUE if the usage is active, FALSE if not.
   */
  private static function isActive(string $entity_type_id, string|int $entity_id, string|int $entity_vid, string $langcode): bool {
    $is_active = static::getActiveRevisionId($entity_type_id, $entity_id, $langcode) === (int) $entity_vid;
    // For a paragraph to be considered active is must also be used by the
    // latest revision of its host entity.
    if ($is_active && $entity_type_id === 'paragraph') {
      $entityTypeManager = \Drupal::entityTypeManager();
      $storage = $entityTypeManager->getStorage('paragraph');
      assert($storage instanceof RevisionableStorageInterface);
      $vid = $storage->getLatestRevisionId($entity_id);
      $entity = $storage->loadRevision($vid);
      while ($entity->getEntityTypeId() === 'paragraph') {
        assert($entity instanceof ParagraphInterface);
        $parent_entity_type_id = $entity->get('parent_type')->value;
        $parent_entity_id = $entity->get('parent_id')->value;
        if ($parent_entity_id === NULL || $parent_entity_type_id === NULL) {
          // Cannot determine host entity therefore have to return TRUE.
          return TRUE;
        }
        try {
          $parent_storage = $entityTypeManager->getStorage($parent_entity_type_id);
          if ($entityTypeManager->getDefinition($parent_entity_type_id)->isRevisionable() && $parent_storage instanceof RevisionableStorageInterface) {
            $host_entity_vid = static::getActiveRevisionId($parent_entity_type_id, $parent_entity_id, $langcode);
            $host_entity = $parent_storage->loadRevision($host_entity_vid);
          }
          else {
            $host_entity = $parent_storage->load($parent_entity_id);
          }
        }
        catch (\RuntimeException) {
        }
        // The host no longer exists therefore it cannot be used.
        if (!isset($host_entity)) {
          return FALSE;
        }
        assert($host_entity instanceof FieldableEntityInterface);
        $field_item_list = $host_entity->get($entity->get('parent_field_name')->value);
        $found = FALSE;
        foreach ($field_item_list as $field_item) {
          if ($field_item->entity->id() === $entity->id()) {
            $found = TRUE;
            break;
          }
        }
        if (!$found) {
          // The paragraph is not used in the default revision.
          return FALSE;
        }
        $entity = $host_entity;
      }
    }
    return $is_active;
  }

  /**
   * Moderates an entity.
   *
   * @param string $entity_type_id
   *   The entity type of the entity to be moderated.
   * @param string|int $entity_id
   *   The entity ID of the entity to be moderated.
   * @param string $moderation_state
   *   The state to change the entity to.
   * @param array $context
   *   The batch context.
   *
   * @return void
   */
  public static function moderateEntity(string $entity_type_id, string|int $entity_id, string $moderation_state, array &$context): void {
    $entity = \Drupal::entityTypeManager()->getStorage($entity_type_id)->load($entity_id);
    if (!$entity) {
      $context['results']['errors'][] = t("Couldn't load @entity_type @id", [
        '@entity_type_id' => $entity_type_id,
        '@id' => $entity_id,
      ]);
      return;
    }

    if (!$entity->access('update')) {
      \Drupal::messenger()->addStatus(t('Denied access to unpublish %label.', ['%label' => $entity->label() ?? 'Unknown']));
      return;
    }

    /** @var ?\Drupal\content_moderation\ModerationInformationInterface $moderation_info */
    $moderation_info = \Drupal::hasService('content_moderation.moderation_information') ? \Drupal::service('content_moderation.moderation_information') : NULL;
    if (!$moderation_info?->isModeratedEntity($entity)) {
      if (!$entity instanceof EntityPublishedInterface) {
        return;
      }

      $current_state = $entity->isPublished() ? 'published' : 'unpublished';
      if ($current_state === $moderation_state) {
        return;
      }

      $revision = static::createRevision($entity);
      assert($revision instanceof EntityPublishedInterface);
      $moderation_state === 'published' ? $revision->setPublished() : $revision->setUnpublished();
      $moderation_state_label = $moderation_state === 'published' ? t('Published') : t('Unpublished');
    }
    else {
      $current_state = $entity->moderation_state->value;
      if ($current_state === $moderation_state) {
        return;
      }
      $revision = static::createRevision($entity);
      assert($revision instanceof ContentEntityInterface);
      $revision->set('moderation_state', $moderation_state);
      $workflow = $moderation_info->getWorkflowForEntity($revision);
      $moderation_state_label = $workflow->getTypePlugin()->getState($moderation_state)->label();
    }

    if ($revision instanceof RevisionLogInterface) {
      $revision->setRevisionLogMessage('Entity usage updater moderation status change');
    }
    // The entity is not validated here because we are only changing the
    // moderation state which is not the same as changing field values.
    $revision->save();
    $context['results']['moderation_message'] = t("Updated moderation state for %label to %state.", ['%label' => $revision->label() ?: $revision->id(), '%state' => $moderation_state_label]);
  }

}
