<?php

declare(strict_types=1);

namespace Drupal\Tests\entity_usage_updater\Functional;

use Drupal\filter\Entity\FilterFormat;
use Drupal\Tests\content_moderation\Functional\ModerationStateTestBase;

/**
 * Tests the Link Remover form with future revisions.
 *
 * @group entity_usage_updater
 */
class FutureRevisionsLinkRemoverFormTest extends ModerationStateTestBase {

  /**
   * The field to add links to.
   */
  const TEXT_FIELD = 'body[0][value]';

  /**
   * Regex to determine if link remover has updated content.
   */
  const UPDATE_MESSAGE_REGEX = '/(updated|Updated)/';

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'node',
    'field_ui',
    'text',
    'path',
    'entity_usage_updater',
    'content_moderation',
  ];

  /**
   * {@inheritdoc}
   */
  public function setUp(): void {
    $this->permissions[] = 'update referenced entities';
    $this->permissions[] = 'administer site configuration';
    parent::setUp();

    $this->drupalLogin($this->adminUser);
    $this->createContentTypeFromUi('Page', 'page', TRUE);
    $this->grantUserPermissionToCreateContentOfType($this->adminUser, 'page');

    // Set up the filter formats used by this test.
    $basic_html_format = FilterFormat::create([
      'format' => 'basic_html',
      'name' => 'Basic HTML',
      'filters' => [
        'filter_html' => [
          'status' => 1,
          'settings' => [
            'allowed_html' => '<p> <br> <strong> <a href> <em>',
          ],
        ],
      ],
    ]);
    $basic_html_format->save();
    user_role_grant_permissions('authenticated', [$basic_html_format->getPermissionName(), 'access content']);

    $current_request = \Drupal::request();
    $config = \Drupal::configFactory()->getEditable('entity_usage.settings');
    $config->set('site_domains', [$current_request->getHttpHost() . $current_request->getBasePath()]);
    $config->save();
    $this->config('entity_usage.settings')
      ->set('local_task_enabled_entity_types', ['node'])
      ->save();
    \Drupal::service('router.builder')->rebuild();
  }

  /**
   * Tests the link remover with a link in both the draft and published content.
   */
  public function testFutureRevisionLinkRemoverBothDraftAndPublished(): void {
    [$node1, $node2, $node3, $node4] = $this->createContent();

    $this->drupalGet('admin/config/content/link-remover');
    $this->assertSession()->fieldExists('links')->setValue(implode("\n", [
      $node1->toUrl()->setAbsolute()->toString(),
      $node2->toUrl()->setAbsolute()->toString(),
    ]));
    $this->assertSession()->buttonExists('Remove')->press();

    $this->assertSession()->buttonExists('Submit')->press();
    // Process the batch.
    $this->checkForMetaRefresh();
    $this->assertSession()->pageTextContains('Updated one entity for references to Node 1.');
    $this->assertSession()->pageTextContains('The latest revision of Node 4 has been updated. The published version is unchanged.');
    // Make an assertion opposite of the one in
    // ::testFutureRevisionLinkRemoverOnlyPublished() just to ensure it works,
    $this->assertSession()->pageTextMatches(static::UPDATE_MESSAGE_REGEX);

    $this->clickLink($node4->label());
    $this->assertSession()->addressMatches('#/node/4/edit$#');

    // We expect one more revision.
    $this->assertSame(3, \Drupal::entityTypeManager()->getStorage('node')->getQuery()->allRevisions()->accessCheck(FALSE)->condition('nid', 4)->count()->execute());

    // Check to see if the link have been removed.
    $this->drupalGet('node/4');
    $this->assertSession()->linkByHrefExists($node1->toUrl()->toString());
    $this->assertSession()->linkByHrefExists($node2->toUrl()->toString());
    $this->assertSession()->linkByHrefNotExists($node3->toUrl()->toString());
    $this->drupalGet('node/4/latest');
    $this->assertSession()->linkByHrefNotExists($node1->toUrl()->toString());
    $this->assertSession()->linkByHrefNotExists($node2->toUrl()->toString());
    $this->assertSession()->linkByHrefExists($node3->toUrl()->toString());
  }

  /**
   * Tests the link remover with a link in only the published content.
   */
  public function testFutureRevisionLinkRemoverOnlyPublished(): void {
    [$node1, $node2, $node3, $node4] = $this->createContent();

    $this->drupalGet('admin/config/content/link-remover');
    $this->assertSession()->fieldExists('links')->setValue(implode("\n", [
      $node2->toUrl()->setAbsolute()->toString(),
    ]));
    $this->assertSession()->buttonExists('Remove')->press();

    $this->assertSession()->buttonExists('Submit')->press();
    // Process the batch.
    $this->checkForMetaRefresh();
    $this->assertSession()->pageTextNotMatches(static::UPDATE_MESSAGE_REGEX);
    $this->assertSession()->linkNotExists($node4->label());

    $this->assertSame(2, \Drupal::entityTypeManager()->getStorage('node')->getQuery()->allRevisions()->accessCheck(FALSE)->condition('nid', 4)->count()->execute());

    // Check to see if the link has not been removed.
    $this->drupalGet('node/4');
    $this->assertSession()->linkByHrefExists($node1->toUrl()->toString());
    $this->assertSession()->linkByHrefExists($node2->toUrl()->toString());
    $this->assertSession()->linkByHrefNotExists($node3->toUrl()->toString());
    $this->drupalGet('node/4/latest');
    $this->assertSession()->linkByHrefExists($node1->toUrl()->toString());
    $this->assertSession()->linkByHrefNotExists($node2->toUrl()->toString());
    $this->assertSession()->linkByHrefExists($node3->toUrl()->toString());
  }

  /**
   * Tests the link remover with a link in only the draft content.
   */
  public function testFutureRevisionLinkRemoverOnlyDraft(): void {
    [$node1, $node2, $node3, $node4] = $this->createContent();

    $this->drupalGet('admin/config/content/link-remover');
    $this->assertSession()->fieldExists('links')->setValue(implode("\n", [
      $node3->toUrl()->setAbsolute()->toString(),
    ]));
    $this->assertSession()->buttonExists('Remove')->press();

    $this->assertSession()->buttonExists('Submit')->press();
    // Process the batch.
    $this->checkForMetaRefresh();
    $this->assertSession()->pageTextContains('Updated one entity for references to Node 3.');
    $this->assertSession()->pageTextContains('The latest revision of Node 4 has been updated. The published version is unchanged.');
    $this->clickLink($node4->label());
    $this->assertSession()->addressMatches('#/node/4/edit$#');

    $this->assertSame(3, \Drupal::entityTypeManager()->getStorage('node')->getQuery()->allRevisions()->accessCheck(FALSE)->condition('nid', 4)->count()->execute());

    // Check to see if the link has not been removed.
    $this->drupalGet('node/4');
    $this->assertSession()->linkByHrefExists($node1->toUrl()->toString());
    $this->assertSession()->linkByHrefExists($node2->toUrl()->toString());
    $this->assertSession()->linkByHrefNotExists($node3->toUrl()->toString());
    $this->drupalGet('node/4/latest');
    $this->assertSession()->linkByHrefExists($node1->toUrl()->toString());
    $this->assertSession()->linkByHrefNotExists($node2->toUrl()->toString());
    $this->assertSession()->linkByHrefNotExists($node3->toUrl()->toString());
  }

  /**
   * Creates content for testing.
   *
   * @return \Drupal\node\NodeInterface[]
   *   The created content in an array.
   */
  private function createContent(): array {
    $session = $this->getSession();
    $page = $session->getPage();
    $assert_session = $this->assertSession();

    // Create node 1.
    $this->drupalGet('node/add/page');
    $page->fillField('title[0][value]', 'Node 1');
    $page->selectFieldOption('moderation_state[0][state]', 'published');
    $page->pressButton('Save');
    $assert_session->pageTextContains('Node 1 has been created.');
    /** @var \Drupal\node\NodeInterface $node1 */
    $node1 = $this->drupalGetNodeByTitle('Node 1');

    // Create node 2.
    $this->drupalGet('node/add/page');
    $page->fillField('title[0][value]', 'Node 2');
    $page->selectFieldOption('moderation_state[0][state]', 'published');
    $page->pressButton('Save');
    $assert_session->pageTextContains('Node 2 has been created.');
    /** @var \Drupal\node\NodeInterface $node2 */
    $node2 = $this->drupalGetNodeByTitle('Node 2');

    // Create node 3.
    $this->drupalGet('node/add/page');
    $page->fillField('title[0][value]', 'Node 3');
    $page->selectFieldOption('moderation_state[0][state]', 'published');
    $page->pressButton('Save');
    $assert_session->pageTextContains('Node 3 has been created.');
    /** @var \Drupal\node\NodeInterface $node3 */
    $node3 = $this->drupalGetNodeByTitle('Node 3');

    // Create node 4.
    $this->drupalGet('node/add/page');
    $page->fillField('title[0][value]', 'Node 4');
    $page->fillField(static::TEXT_FIELD, $node1->toLink("Link", options: ['absolute' => TRUE])->toString() . $node2->toLink("Link", options: ['absolute' => TRUE])->toString());
    $page->selectFieldOption('moderation_state[0][state]', 'published');
    $page->pressButton('Save');
    $assert_session->pageTextContains('Node 4 has been created.');

    // Create future draft node 4.
    $this->drupalGet('node/4/edit');
    $page->fillField(static::TEXT_FIELD, $node1->toLink("Link", options: ['absolute' => TRUE])->toString() . $node3->toLink("Link", options: ['absolute' => TRUE])->toString());
    $page->selectFieldOption('moderation_state[0][state]', 'draft');
    $page->pressButton('Save');
    $assert_session->pageTextContains('Node 4 has been updated.');
    $node4 = $this->drupalGetNodeByTitle('Node 4');

    // Ensure the links are as expected before using the link remover.
    $this->drupalGet('node/4');
    $this->assertSession()->linkByHrefExists($node1->toUrl()->toString());
    $this->assertSession()->linkByHrefExists($node2->toUrl()->toString());
    $this->assertSession()->linkByHrefNotExists($node3->toUrl()->toString());
    $this->drupalGet('node/4/latest');
    $this->assertSession()->linkByHrefExists($node1->toUrl()->toString());
    $this->assertSession()->linkByHrefNotExists($node2->toUrl()->toString());
    $this->assertSession()->linkByHrefExists($node3->toUrl()->toString());

    return [$node1, $node2, $node3, $node4];
  }

}
